﻿/**

*/
module nemuxi.file.items.items;

debug import std.stdio: wl = writefln, pl = printf;
debug(items) void main() {}

import str = std.string;
import std.utf;

import std.contracts;

import etc.kareki.kareki;

import nemuxi.base;
import nemuxi.system.timer;
import nemuxi.file.items.item;
import nemuxi.file.items.linkitem;
import nemuxi.file.items.itemfunc;
import nemuxi.system.application;
import nemuxi.image.icon;
import nemuxi.system.type;

/**
アイテム・アイコンペア。

<ul>
	<li>L:アイテム</li>
	<li>R:アイコン</li>
</ul>
*/
alias PAIR!(Item, Icon) ITEMICONPAIR;

/**
アイテム・アイコンペア統括。

推奨用途は読み取り。
*/
final class ItemIconPairs {
	immutable ICONSIZE iconSize;
	const int iconPx() {
		return GetIconSize(iconSize);
	}
	
	private {
		size_t[wstring] index;
		ITEMICONPAIR[] pairs;
	}

	this(Kareha ItemTree, ICONSIZE IconSize) {
		iconSize = IconSize;

		pairs = new ITEMICONPAIR[ItemTree.length];
		
		foreach(i, key; ItemTree.keys) {
			const ItemID=key.toText;
			index[ItemID.text] = i;
			pairs[i].left  = new Item(ItemID, ItemTree);
			pairs[i].right = GetIcon(pairs[i].left, IconSize, ItemTree);
		}
	}
	~this() {
		if(pairs) {
			foreach(pair; pairs) {
				delete pair.left;
			}
			delete pairs;
		}
	}
	


	const size_t length() {
		return pairs.length;
	}

	size_t keyToIndex(in Text Key) {
		if(auto pIndex = (Key.text in index)) {
			return *pIndex;
		} else {
			throw new NemuxiException("あ～あ");
		}
	}

	ITEMICONPAIR* opIn_r(size_t Index) {
		if(0 <= Index && Index < pairs.length) {
			return &pairs[Index];
		}

		return null;
	}
	ITEMICONPAIR* opIn_r(in Text Key) {
		if(auto pIndex = (Key.text in index)) {
			return opIn_r(*pIndex);
		}
		return null;
	}
	ref ITEMICONPAIR opIndex(size_t Index) {
		return pairs[Index];
	}
	ref ITEMICONPAIR opIndex(in Text Key) {
		return pairs[*(Key.text in index)];
	}

	void addItem(Item item, Kareha ItemTree) {
		ITEMICONPAIR Pair;
		Pair.left = item;
		Pair.right= GetIcon(item, iconSize, ItemTree);

		index[item.itemID.text] = pairs.length;
		pairs ~= Pair;
	}
	
	void changeItem(in Text SrcID, Item NewItem, Kareha ItemTree)
	in {
		if(SrcID != NewItem.itemID)
		assert((NewItem.itemID.text in index) is null);
	}
	body {
		auto Index=*(SrcID.text in index);
		index.remove(cast(immutable)SrcID.text);
		assert((SrcID.text in index) is null, "aaaa");
		index[NewItem.itemID.text] = Index;
		pairs[Index].left  = NewItem;
		pairs[Index].right = GetIcon(NewItem, iconSize, ItemTree);
	}

	bool del(in Text ItemID) {
		foreach(i; 0 .. pairs.length) {
			if(pairs[i].left.itemID == ItemID) {
				if(pairs.length != 1) {
					if(i) {
						pairs = pairs[0..i] ~ pairs[i+1 .. pairs.length];
					} else {
						pairs = pairs[1 .. pairs.length];
					}
					
					foreach(j; 0 .. pairs.length) {
						index[pairs[j].left.itemID.text] = j;
					}
				} else {
					pairs = null;
					index = null;
				}

				return true;
			}
		}

		return false;
	}

}
debug(items) unittest {
	auto a=MakeApplication();

	auto iip=new ItemIconPairs(a.itemBase, ICONSIZE.SMALL);


}


/**
アイテムとアイコンのリスト。
wchar-wstringを信じる。

Bugs:
	ItemIconPairと統合すべし。
*/
class ItemList {
	/+
	invariant() {
		if(!ShowIcon)
			assert(!icons);
	}
	+/
	private {
		Icon[wstring] icons;
		//bool ShowIcon;
		ICONSIZE IconSize;
		Tree tree;
	}
	private class Tree {
		Tree[wchar] Next;
		CITEM[]    CItems;

		~this() {
			if(Next) {
				scope Nexts=new Tree[Next.length];
				foreach(i, key; Next.keys) {
					Nexts[i] = Next[key];
				}
				Next = null;
				foreach(next; Nexts) {
					delete next;
				}
			}
			CItems = null;
		}

		/// 長！
		private bool Set(CITEM.TYPE Type, Item item, size_t Index, string s, out wchar Key, ref Item[] NextItems, lazy Kareha ItemBase=null)
		in {
			assert(item);
			assert(s.length);
		}
		body {
			
			auto value=toUTF32(s);
			if(value.length > Index) {
				CITEM CItem;
				CItem.item = item;
				CItem.type = Type;
				CItem.typeString = s;

				bool copy=true;
				foreach(ref ci; CItems) {
					if(ci == CItem) {
						copy=false;
						break;
					}
				}
				if(copy) {
					if(!(CItem.item.itemID.idup in icons)) {
						assert(ItemBase);
						icons[CItem.item.itemID.idup] = GetIcon(CItem.item, IconSize, ItemBase);
					}
					CItem.icon = icons[CItem.item.itemID.idup];
					CItems ~= CItem;
				}
				if(value.length != Index) {
					NextItems ~= item;
					Key = cast(wchar)value[Index];
					enforce(Key == value[Index], new NemuxiException(Text("index-"d ~ value[Index])));
				}
				
				return true;
			}

			return false;
		}
		void set(CITEM.TYPE Type, Item[] ItemList, size_t Index, lazy Kareha ItemBase=null)
		in {
			assert(ItemList.length);
		}
		body {
			foreach(item; ItemList) {
				Item[] NextItems;
				CITEM CItem;
				wchar Key;
				wchar[] Keys;
				switch(Type) {
					case CITEM.TYPE.ID:
						Set(Type, item, Index, item.itemID.toString, Key, NextItems, ItemBase);
						break;
					case CITEM.TYPE.NAME:
						if(item.name().length) {
							if(item.itemID != item.name())
							Set(Type, item, Index, item.name.toString, Key, NextItems, ItemBase);
						}
						break;
					case CITEM.TYPE.TAG: {
						auto tags=item.tags();
						Keys = new wchar[tags.length];
						foreach(i, tag; tags) {
							Set(Type, item, Index, tag.toString, Keys[i], NextItems, ItemBase);
						}
						break;
					}
					default:
						assert(false);
				}
				if(NextItems) {
					if(Keys) {
						assert(Key == Key.init);
						foreach(key; Keys) {
							if(!(key in Next)) {
								Next[key] = new Tree();
							}
							Next[key].set(Type, NextItems, Index+1, ItemBase);
						}
					} else {
						assert(Key != Key.init);
						if(!(Key in Next)) {
							Next[Key] = new Tree();
						}
						Next[Key].set(Type, NextItems, Index+1, ItemBase);
					}
				}
			}
		}
	}
	
	this(Kareha ItemBase, ICONSIZE IconSize)
	in {
		assert(ItemBase);
	}
	body {
		this.IconSize = IconSize;
		auto items=new Item[ItemBase.length];
		tree = new Tree();
		foreach(i, key; ItemBase.keys) {
			items[i] = new Item(key.toText, ItemBase);
		}
		for(CITEM.TYPE i=CITEM.TYPE.min; i < CITEM.TYPE.max+1; i++) {
			tree.set(i, items, 0, ItemBase);
		}
	}
	
	~this() {
		delete tree;
		if(icons) {
			scope Icons=new Icon[icons.length];
			foreach(i, key; icons.keys) {
				Icons[i] = icons[key];
			}
			icons = null;
			foreach(icon; Icons) {
				delete icon;
			}
		}
	}

	CITEM[] find(wstring s, bool Exactness=false) {
		Tree Loop=tree;
		
		foreach(i, c; s) {
			if(c in Loop.Next) {
				if(s.length-1 > i) {
					Loop = Loop.Next[c];
				} else if(s.length-1 == i) {
					return Loop.Next[c].CItems;
				}
			} else {
				return null;
			}
		}

		return tree.CItems;
	}
	CITEM[] find(wstring s, CITEM.TYPE Type) {
		auto CItems=this.find(s);
		
		if(!CItems.length) {
			return null;
		}
		
		CITEM[] List;
		foreach(ref CItem; CItems) {
			if(CItem.type == Type) {
				List ~= CItem;
			}
		}

		return List;
	}
	bool inIcon(string key) {
		return cast(bool)(key.toText.idup in icons);
	}
	Icon icon(string key) {
		assert(key.length);
		return icons[key.toText.idup];
	}
	ICONSIZE iconSize() {
		return IconSize;
	}
	int iconPx() {
		return GetIconSize(this.iconSize());
	}

	//CITEM
	/+
	debug(items) unittest {
		auto a=MakeApplication();
		//auto show= a.AppData.exeCommandIconUse;
		scope dt=new DTimer;
		//foreach(n; 1..300){
			scope l=new ItemList(a.itemBase(), a.AppData.exeCommandIcon());
			
			wl("[i]----------------------------");
			foreach(c; l.find("i")){
				wl("%s",c);
			}
			wl("[i_]---------------------------");
			foreach(c; l.find("i_")){
				wl("%s",c);
			}
			wl("[iii]---------------------------");
			foreach(c; l.find("ii")){
				wl("%s",c);
			}
			//wl("%s", l.tree.Next['i'].Next['_'].CItems);
		//}
	};
	+/
}


